
{*******************************************************}
{                                                       }
{       Fetisov Sergey Visual Component Library         }
{                for graphic working                    }
{                    14.02.2003                         }
{                                                       }
{*******************************************************}

unit GraphControls;
                         
interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Animation,
  Graphics;

{$R *.res}

type
  TGraphicButtonImages = class(TPersistent)
  private
    FNormalImg: TBitmap;
    FMovingImg: TBitmap;
    FDisabledImg: TBitmap;
    FDownImg: TBitmap;
    procedure SetDisabledImg(const Value: TBitmap);
    procedure SetDownImg(const Value: TBitmap);
    procedure SetMovingImg(const Value: TBitmap);
    procedure SetNormalImg(const Value: TBitmap);
  protected
    OnChangeImg: TNotifyEvent;
    procedure Change(Sender: TObject);
  public
    destructor Destroy; override;
    constructor Create; virtual;
    procedure SaveToStream(Stream: TStream); virtual;
    procedure SaveToFile(FileName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure LoadFromFile(FileName: string); virtual;
  published
    property NormalImg: TBitmap read FNormalImg write SetNormalImg;
    property MovingImg: TBitmap read FMovingImg write SetMovingImg;
    property DownImg: TBitmap read FDownImg write SetDownImg;
    property DisabledImg: TBitmap read FDisabledImg write SetDisabledImg;
  end;

  TGraphicButtonState = (gsNormal, gsMove, gsDown);

  TCustomGraphicButton = class(TGraphicControl)
  private
    FImages: TGraphicButtonImages;
    FOnMouseLeave: TNotifyEvent;
    FOnMouseEnter: TNotifyEvent;
    FTransparent: Boolean;
    procedure SetImages(const Value: TGraphicButtonImages);
    procedure SetTransparent(const Value: Boolean);
  protected
    BtnState: TGraphicButtonState;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property Images: TGraphicButtonImages read FImages write SetImages;
    property Transparent: Boolean read FTransparent write SetTransparent;
    procedure ImagesChange(Sender: TObject); virtual;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X: Integer; Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X: Integer;
      Y: Integer); override;
    procedure CMFontchanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMTextchanged(var Message: TMessage); message CM_TEXTCHANGED;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  TGraphicButton = class(TCustomGraphicButton)
  published
    property Transparent;
    property Images;
    property Color;
    property Caption;
    property Enabled;
    property Font;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnMouseEnter;
    property OnMouseLeave;
  end;

  TDisplayItem = class(TPersistent)
  private
    FOnChange: TNotifyEvent;
    FView: TBitmap;
    FName: string;
    FList: Pointer;
    FLeft: Integer;
    FTop: Integer;
    FHeight: Integer;
    FWidth: Integer;
    FAutoSize: Boolean;
    FCanCallChange: Boolean;
    procedure SetView(const Value: TBitmap); virtual;
    procedure SetLeft(const Value: Integer); virtual;
    procedure SetTop(const Value: Integer); virtual;
    procedure SetHeight(const Value: Integer); virtual;
    procedure SetWidth(const Value: Integer); virtual;
    procedure ViewChange(Sender: TObject); virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Change; virtual;
    procedure Move(DeltaX, DeltaY: Integer); virtual;
    function GetRect: TRect;
    property List: Pointer read FList write FList;
  published
    property CanCallChange: Boolean read FCanCallChange write FCanCallChange;
    property AutoSize: Boolean read FAutoSize write FAutoSize;
    property Height: Integer read FHeight write SetHeight;
    property Left: Integer read FLeft write SetLeft;
    property Name: string read FName write FName;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Top: Integer read FTop write SetTop;
    property View: TBitmap read FView write SetView;
    property Width: Integer read FWidth write SetWidth;
  end;

  TAnimatedDisplayItem = class(TDisplayItem)
  private
    FAnimation: TAnimation;
  public
    constructor Create; override;
    destructor Destroy; override;
    property Animation: TAnimation read FAnimation write FAnimation;
  protected
    procedure FrameChange(Sender: TCustomAnimation; Frame: TAnimationFrame); virtual;
    procedure BeginPlay(Sender: TCustomAnimation; Frame: TAnimationFrame); virtual;
    procedure EndPlay(Sender: TCustomAnimation; Frame: TAnimationFrame); virtual;
  published
    property AutoSize;
    property CanCallChange;
    property Height;
    property Left;
    property Name;
    property OnChange;
    property Top;
    property View;
    property Width;
  end;

  TObjMouseMovingAccept = procedure(Sender: TObject; DisplayObj: TDisplayItem; var Accept: Boolean)of object;
  TObjMouseEvent = procedure(Sender: TObject; DisplayObj: TDisplayItem; Button: TMouseButton; Shift: TShiftState; X, Y: Integer)of object;
  TObjMouseMoveEvent = procedure(Sender: TObject; DisplayObj: TDisplayItem; Shift: TShiftState; X, Y: Integer)of object;
  TObjChangeEvent = procedure(Sender: TObject; Obj: TDisplayItem)of object;

  TDisplayItems = class(TPersistent)
  private
    List: TList;
    FOnCallChange: TNotifyEvent;
    function GetCount: Integer;
    function GetItems(Index: Integer): TDisplayItem;
    procedure SetItems(Index: Integer; const Value: TDisplayItem);
    procedure SetOnCallChange(const Value: TNotifyEvent);
    procedure DoChange;
  public
    constructor Create;
    destructor Destroy; override;
    property Count: Integer read GetCount;
    property Items[Index: Integer]: TDisplayItem read GetItems write SetItems; default;
    property OnCallChange: TNotifyEvent read FOnCallChange write SetOnCallChange;
    function ItemExists(Index: Integer): Boolean;
    procedure Add(NewItem: TDisplayItem);
    procedure Delete(Index: Integer);
    procedure Move(FirstIndex, SecondIndex: Integer);
    function IndexOf(Item: TDisplayItem): Integer;
    function IndexOfName(ItemName: string): Integer;
  end;

  TDisplay = class(TGraphicControl)
  private
    DisplayImage: TBitmap;
    FAutoDraw: Boolean;
    FDrawFirstOnTop: Boolean;
    FAutoObjMouseMoving: Boolean;
    FOnObjMouseMovingAccept: TObjMouseMovingAccept;
    FOnObjMouseUp: TObjMouseEvent;
    FOnObjMouseDown: TObjMouseEvent;
    FOnObjMouseMove: TObjMouseMoveEvent;
    Moving: Boolean;
    MovingPt: TPoint;
    MovingObj: TDisplayItem;
    FOnObjChange: TObjChangeEvent;
    FItems: TDisplayItems;
    procedure ChangeItem(Sender: TObject); virtual;
    procedure SetDrawFirstOnTop(const Value: Boolean);
    procedure SetItems(const Value: TDisplayItems);
  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X: Integer; Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X: Integer;
      Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X: Integer; Y: Integer);
      override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Items: TDisplayItems read FItems write SetItems;
    procedure DrawDisplay; virtual;
    procedure GetDisplayImage(var BitmapImage: TBitmap); virtual;
    procedure MoveAllItems(DeltaX, DeltaY: Integer); virtual;
    function FindObjAtPos(X, Y: Integer): TDisplayItem;
  published
    property AutoDraw: Boolean read FAutoDraw write FAutoDraw;
    property DrawFirstOnTop: Boolean read FDrawFirstOnTop write SetDrawFirstOnTop;
    property AutoObjMouseMoving: Boolean read FAutoObjMouseMoving write FAutoObjMouseMoving;
    property OnObjMMovingAccept: TObjMouseMovingAccept read FOnObjMouseMovingAccept write FOnObjMouseMovingAccept;
    property OnObjMDown: TObjMouseEvent read FOnObjMouseDown write FOnObjMouseDown;
    property OnObjMUp: TObjMouseEvent read FOnObjMouseUp write FOnObjMouseUp;
    property OnObjMMove: TObjMouseMoveEvent read FOnObjMouseMove write FOnObjMouseMove;
    property OnObjChange: TObjChangeEvent read FOnObjChange write FOnObjChange;
    property OnCanResize;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property Color;
    property Align;
    property ParentColor;
    property ParentShowHint;
  end;

  TDrawImgStyle = (diLeftTop, diCenter, diStretch);
  TDrawSectorEvent = procedure(Sender: TObject; ACanvas: TCanvas; Sector: TPoint; ARect: TRect) of object;
  TSectorMouseEvent = procedure(Sender: TObject; Button: TMouseButton; Shift: TShiftState; SectorX, SectorY: Integer) of object;
  TSectorMMoveEvent = procedure(Sender: TObject; Shift: TShiftState; SectorX, SectorY: Integer) of object;

  TSectorField = class(TGraphicControl)
  private
    Img: TBitmap;
    SectorsImg: TBitmap;
    FBackImage: TBitmap;
    FHorSectorCount: Integer;
    FHorSectorSize: Integer;
    FVertSectorCount: Integer;
    FVertSectorSize: Integer;
    FDrawImageStyle: TDrawImgStyle;
    FOnDrawSector: TDrawSectorEvent;
    FPen: TPen;
    FOnSectorUp: TSectorMouseEvent;
    FOnSectorDown: TSectorMouseEvent;
    FOnSectorMove: TSectorMMoveEvent;
    procedure Size;
    procedure SetBackImage(const Value: TBitmap);
    procedure SetHorSectorCount(const Value: Integer);
    procedure SetHorSectorSize(const Value: Integer);
    procedure SetVertSectorCount(const Value: Integer);
    procedure SetVertSectorSize(const Value: Integer);
    procedure SetDrawImageStyle(const Value: TDrawImgStyle);
    procedure SetPen(const Value: TPen);
  protected
    procedure Paint; override;
    procedure PenChange(Sender: TObject);
    procedure MouseMove(Shift: TShiftState; X: Integer; Y: Integer);
      override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X: Integer; Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X: Integer;
      Y: Integer); override;
    procedure SetAutoSize(Value: Boolean); override;
    procedure Resize; override;
    procedure DrawSectorsImg;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetSectorRect(Sector: TPoint): TRect;
    function SectorAtPos(X, Y: Integer): TPoint;
    procedure Redraw;
    procedure DrawSectors;
  published
    property HorSectorSize: Integer read FHorSectorSize write SetHorSectorSize;
    property HorSectorCount: Integer read FHorSectorCount write SetHorSectorCount;
    property VertSectorSize: Integer read FVertSectorSize write SetVertSectorSize;
    property VertSectorCount: Integer read FVertSectorCount write SetVertSectorCount;
    property BackImage: TBitmap read FBackImage write SetBackImage;
    property Pen: TPen read FPen write SetPen;
    property DrawImageStyle: TDrawImgStyle read FDrawImageStyle write SetDrawImageStyle;
    property AutoSize;
    property OnDrawSector: TDrawSectorEvent read FOnDrawSector write FOnDrawSector;
    property OnSectorMove: TSectorMMoveEvent read FOnSectorMove write FOnSectorMove;
    property OnSectorDown: TSectorMouseEvent read FOnSectorDown write FOnSectorDown;
    property OnSectorUp: TSectorMouseEvent read FOnSectorUp write FOnSectorUp;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property Color;
    property ParentColor;
    property ParentShowHint;
    property ShowHint;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Graph controls', [TGraphicButton, TDisplay, TSectorField]);
end;

{ TGraphicButtonImages }

procedure TGraphicButtonImages.Change(Sender: TObject);
begin
  if Assigned(OnChangeImg)then
    OnChangeImg(Sender);
end;

constructor TGraphicButtonImages.Create;
begin
  FNormalImg:=TBitmap.Create;
  FMovingImg:=TBitmap.Create;
  FDownImg:=TBitmap.Create;
  FDisabledImg:=TBitmap.Create;

  FNormalImg.OnChange:=Change;
  FMovingImg.OnChange:=Change;
  FDownImg.OnChange:=Change;
  FDisabledImg.OnChange:=Change;
end;

destructor TGraphicButtonImages.Destroy;
begin
  FNormalImg.Free;
  FMovingImg.Free;
  FDisabledImg.Free;
  FDownImg.Free;
  inherited;
end;

procedure TGraphicButtonImages.LoadFromFile(FileName: string);
var
  Stream: TFileStream;
begin
  Stream:=TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    if Stream<>nil then
      Stream.Free;
  end
end;

procedure TGraphicButtonImages.LoadFromStream(Stream: TStream);
begin
  FNormalImg.LoadFromStream(Stream);
  FMovingImg.LoadFromStream(Stream);
  FDownImg.LoadFromStream(Stream);
  FDisabledImg.LoadFromStream(Stream);
end;

procedure TGraphicButtonImages.SaveToFile(FileName: string);
var
  Stream: TFileStream;
begin
  Stream:=TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    if Stream<>nil then
      Stream.Free;
  end
end;

procedure TGraphicButtonImages.SaveToStream(Stream: TStream);
begin
  FNormalImg.SaveToStream(Stream);
  FMovingImg.SaveToStream(Stream);
  FDownImg.SaveToStream(Stream);
  FDisabledImg.SaveToStream(Stream);
end;

procedure TGraphicButtonImages.SetDisabledImg(const Value: TBitmap);
begin
  FDisabledImg.Assign(Value);
  Change(FDisabledImg);
end;

procedure TGraphicButtonImages.SetDownImg(const Value: TBitmap);
begin
  FDownImg.Assign(Value);
  Change(FDownImg);
end;

procedure TGraphicButtonImages.SetMovingImg(const Value: TBitmap);
begin
  FMovingImg.Assign(Value);
  Change(FMovingImg);
end;

procedure TGraphicButtonImages.SetNormalImg(const Value: TBitmap);
begin
  FNormalImg.Assign(Value);
  Change(FNormalImg);
end;

{ TCustomGraphicButton }

procedure TCustomGraphicButton.CMFontchanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TCustomGraphicButton.CMMouseEnter(var Message: TMessage);
begin
  if BtnState=gsNormal then
  begin
    BtnState:=gsMove;
    Invalidate;
  end;
end;

procedure TCustomGraphicButton.CMMouseLeave(var Message: TMessage);
begin
  if BtnState=gsMove then
  begin
    BtnState:=gsNormal;
    Invalidate;
  end;
end;

procedure TCustomGraphicButton.CMTextchanged(var Message: TMessage);
begin
  Invalidate;
end;

constructor TCustomGraphicButton.Create(AOwner: TComponent);
begin
  inherited;
  FImages:=TGraphicButtonImages.Create;
  FImages.OnChangeImg:=ImagesChange;
  BtnState:=gsNormal;
  FTransparent:=false;
end;

destructor TCustomGraphicButton.Destroy;
begin
  FImages.Free;
  inherited;
end;

procedure TCustomGraphicButton.ImagesChange(Sender: TObject);
begin
  if ((Sender=Images.NormalImg)and(BtnState=gsNormal))or
  ((Sender=Images.MovingImg)and(BtnState=gsMove))or
  ((Sender=Images.DownImg)and(BtnState=gsDown))then
    Invalidate;
end;

procedure TCustomGraphicButton.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if Button<>mbLeft then Exit;
  BtnState:=gsDown;
  Invalidate;
end;

procedure TCustomGraphicButton.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if (X>=0)and(X<=Width)and(Y>=0)and(Y<=Height)then
    BtnState:=gsMove else
      BtnState:=gsNormal;
  Invalidate;
end;

procedure TCustomGraphicButton.Paint;
var
  X, Y, W, H: Integer;
begin
  inherited;
  with Canvas do
  begin
    Brush.Color:=Self.Color;
    if not Transparent then
      FillRect(ClipRect);
    if not Enabled then
      Draw(0, 0, FImages.FDisabledImg)else
        case BtnState of
          gsNormal: Draw(0, 0, FImages.FNormalImg);
          gsMove: Draw(0, 0, FImages.FMovingImg);
          gsDown: Draw(0, 0, FImages.FDownImg);
        end;
    Font:=Self.Font;
    W:=TextWidth(Caption);
    H:=TextHeight('0');
    X:=Width div 2 - W div 2;
    Y:=Height div 2 - H div 2;
    TextOut(X, Y, Caption);
  end;
end;

procedure TCustomGraphicButton.SetImages(
  const Value: TGraphicButtonImages);
begin
  FImages.Assign(Value);
  Invalidate;
end;

procedure TCustomGraphicButton.SetTransparent(const Value: Boolean);
begin
  FTransparent := Value;
  if Value then
    ControlStyle:=ControlStyle+[csReplicatable]else
      ControlStyle:=ControlStyle-[csReplicatable];
  FImages.FNormalImg.Transparent:=Value;
  FImages.FMovingImg.Transparent:=Value;
  FImages.FDownImg.Transparent:=Value;
  FImages.FDisabledImg.Transparent:=Value;
  Invalidate;
end;

{ TDisplayItem }

procedure TDisplayItem.Change;
begin
  if AutoSize then
  begin
    FWidth:=View.Width;
    FHeight:=View.Height;
  end;
  if not FCanCallChange then Exit;
  if Assigned(FOnChange)then
    FOnChange(Self);
end;

constructor TDisplayItem.Create;
begin
  FCanCallChange:=false;
  FName:='';
  FLeft:=0;
  FTop:=0;
  FWidth:=0;
  FHeight:=0;
  FAutoSize:=true;
  FView:=TBitmap.Create;
  FView.OnChange:=ViewChange;
  FView.Transparent:=true;
  FView.TransparentColor:=clWhite;
  FCanCallChange:=true;
end;

destructor TDisplayItem.Destroy;
begin
  FView.Free;
  inherited;
end;

function TDisplayItem.GetRect: TRect;
begin
  Result:=Rect(Left, Top, Left+Width, Top+Height);
end;

procedure TDisplayItem.Move(DeltaX, DeltaY: Integer);
var
  OldCanCallChange: Boolean;
begin
  OldCanCallChange:=CanCallChange;
  CanCallChange:=false;
  Left:=Left+DeltaX;
  Top:=Top+DeltaY;
  CanCallChange:=OldCanCallChange;
  Change;
end;

procedure TDisplayItem.SetHeight(const Value: Integer);
begin
  FHeight := Value;
  Change;
end;

procedure TDisplayItem.SetLeft(const Value: Integer);
begin
  FLeft := Value;
  Change;
end;

procedure TDisplayItem.SetTop(const Value: Integer);
begin
  FTop := Value;
  Change;
end;

procedure TDisplayItem.SetView(const Value: TBitmap);
begin
  FView.Assign(Value);
end;

procedure TDisplayItem.SetWidth(const Value: Integer);
begin
  FWidth := Value;
  Change;
end;

procedure TDisplayItem.ViewChange(Sender: TObject);
begin
  Change;
end;

{ TAnimatedDisplayItem }

procedure TAnimatedDisplayItem.BeginPlay(Sender: TCustomAnimation;
  Frame: TAnimationFrame);
begin
//
end;

constructor TAnimatedDisplayItem.Create;
begin
  inherited;
  FAnimation:=TAnimation.Create;
  FAnimation.OnFrameChange:=FrameChange;
  FAnimation.OnBedinPlay:=BeginPlay;
  FAnimation.OnEndPlay:=EndPlay;
end;

destructor TAnimatedDisplayItem.Destroy;
begin
  FAnimation.Free;
  inherited;
end;

procedure TAnimatedDisplayItem.EndPlay(Sender: TCustomAnimation;
  Frame: TAnimationFrame);
begin
//
end;

procedure TAnimatedDisplayItem.FrameChange(Sender: TCustomAnimation;
  Frame: TAnimationFrame);
var
  OldTransp: Boolean;
  OldTranspColor: TColor;
  OldCanCallChange: Boolean;
begin
  if Frame<>nil then
  begin
    OldCanCallChange:=CanCallChange;
    CanCallChange:=false;
    OldTranspColor:=View.TransparentColor;
    OldTransp:=View.Transparent;
    Move(Frame.DeltaX, Frame.DeltaY);
    View.Assign(Frame.Image);
    View.Transparent:=OldTransp;
    View.TransparentColor:=OldTranspColor;
    CanCallChange:=OldCanCallChange;
    Change;
  end;
end;

{ TDisplay }

procedure TDisplay.ChangeItem(Sender: TObject);
begin
  if Sender is TDisplayItem then
    if Assigned(FOnObjChange)then
      FOnObjChange(Self, Sender as TDisplayItem);
  if AutoDraw then
    Paint;
end;

constructor TDisplay.Create(AOwner: TComponent);
begin
  inherited;
  DisplayImage:=TBitmap.Create;
  FItems:=TDisplayItems.Create;
  FItems.FOnCallChange:=ChangeItem;
  Width:=400;
  Height:=200;
  Color:=clBlack;
  FAutoDraw:=false;
  FDrawFirstOnTop:=false;
  FAutoObjMouseMoving:=false;
  Moving:=false;
end;

destructor TDisplay.Destroy;
begin
  DisplayImage.Free;
  FItems.Free;
  inherited Destroy;
end;

procedure TDisplay.DrawDisplay;
begin
  Paint;
end;

function TDisplay.FindObjAtPos(X, Y: Integer): TDisplayItem;
var
  I: Integer;
  Obj: TDisplayItem;
  Pixel: TColor;
begin
  Result:=nil;
  for I:=0 to FItems.Count-1 do
  begin
    Obj:=FItems[I];
    if Obj<>nil then
      if ((X>=Obj.Left)and(X<=Obj.Left+Obj.Width))and
        ((Y>=Obj.Top)and(Y<=Obj.Top+Obj.Height))then
      begin
        Pixel:=Obj.View.Canvas.Pixels[X-Obj.Left, Y-Obj.Top];
        if(not Obj.View.Transparent)or
        (Obj.View.Transparent and(Pixel or $02000000<>Obj.View.TransparentColor))then
        begin
          Result:=Obj;
          if FDrawFirstOnTop and(I=0)then
            Exit;
        end;
      end;
  end;
end;

procedure TDisplay.GetDisplayImage(var BitmapImage: TBitmap);
begin
  if BitmapImage<>nil then
    BitmapImage.Assign(DisplayImage);
end;

procedure TDisplay.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  Obj: TDisplayItem;
begin
  inherited;
  Obj:=FindObjAtPos(X, Y);
  MovingObj:=Obj;
  if Button=mbLeft then
    Moving:=true;
  if Obj<>nil then
  begin
    MovingPt:=Point(X-Obj.Left, Y-Obj.Top);
    if Assigned(FOnObjMouseDown)then
      FOnObjMouseDown(Self, Obj, Button, Shift, X-Obj.Left, Y-Obj.Top);
  end;
end;

procedure TDisplay.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Obj: TDisplayItem;
  Accept: Boolean;
begin
  inherited;
  if Moving and(MovingObj<>nil)then
  begin
    Accept:=FAutoObjMouseMoving;
    if Assigned(FOnObjMouseMovingAccept)then
      FOnObjMouseMovingAccept(Self, MovingObj, Accept);
    if Accept then
    begin
      MovingObj.Left:=X-MovingPt.X;
      MovingObj.Top:=Y-MovingPt.Y;
      if AutoDraw then
        DrawDisplay;
    end;
  end;
  if Assigned(FOnObjMouseMove)then
  begin
    Obj:=FindObjAtPos(X, Y);
    if Obj<>nil then
      FOnObjMouseMove(Self, Obj, Shift, X-Obj.Left, Y-Obj.Top);
  end;
end;

procedure TDisplay.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
var
  Obj: TDisplayItem;
begin
  inherited;
  Moving:=false;
  if Assigned(FOnObjMouseUp)then
  begin
    Obj:=FindObjAtPos(X, Y);
    if Obj<>nil then
      FOnObjMouseUp(Self, Obj, Button, Shift, X-Obj.Left, Y-Obj.Top);
  end;
end;

procedure TDisplay.MoveAllItems(DeltaX, DeltaY: Integer);
var
  I: Integer;
  Obj: TDisplayItem;
  OldAutoDraw: Boolean;
begin
  OldAutoDraw:=FAutoDraw;
  FAutoDraw:=false;
  for I:=0 to FItems.Count-1 do
  begin
    Obj:=FItems[I];
    if Obj<>nil then
      Obj.Move(DeltaX, DeltaY);
  end;
  FAutoDraw:=OldAutoDraw;
  if FAutoDraw then
    DrawDisplay;
end;

procedure TDisplay.Paint;
var
  I: Integer;
  Obj: TDisplayItem;
begin
  inherited;
  if DisplayImage.Width<>Width then
    DisplayImage.Width:=Width;
  if DisplayImage.Height<>Height then
    DisplayImage.Height:=Height;

  with DisplayImage.Canvas do
  begin
    Brush.Color:=Self.Color;
    FillRect(ClipRect);
    if csDesigning in ComponentState then
    begin
      Font.Color:=clWhite;
      if Color=clWhite then
        Font.Color:=clBlack;
      TextOut(5, 5, 'Graphic display');
    end;

    if (FDrawFirstOnTop=false)and(FItems.Count<>0)then
    begin
      Obj:=FItems[0];
      if Obj<>nil then
        Draw(Obj.Left, Obj.Top, Obj.View);
    end;

    for I:=1 to FItems.Count-1 do
    begin
      Obj:=FItems[I];
      if Obj<>nil then
        Draw(Obj.Left, Obj.Top, Obj.View);
    end;

    if FDrawFirstOnTop and(FItems.Count<>0)then
    begin
      Obj:=FItems[0];
      if Obj<>nil then
        Draw(Obj.Left, Obj.Top, Obj.View);
    end;

  end;
  BitBlt(Canvas.Handle, 0, 0, Width, Height, DisplayImage.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TDisplay.SetDrawFirstOnTop(const Value: Boolean);
begin
  if Value=FDrawFirstOnTop then Exit;
  FDrawFirstOnTop := Value;
  if AutoDraw then
    Paint;
end;

{ TSectorField }

constructor TSectorField.Create(AOwner: TComponent);
begin
  inherited;
  Img:=TBitmap.Create;
  SectorsImg:=TBitmap.Create;
  SectorsImg.Transparent:=true;
  SectorsImg.TransparentColor:=1;
  FPen:=TPen.Create;
  FPen.Color:=clNavy;
  FPen.OnChange:=PenChange;
  FBackImage:=TBitmap.Create;
  FVertSectorCount:=10;
  FVertSectorSize:=20;
  FHorSectorCount:=10;
  FHorSectorSize:=20;
  FDrawImageStyle:=diCenter;
  Size;
  DrawSectorsImg;
end;

destructor TSectorField.Destroy;
begin
  FBackImage.Free;
  FPen.Free;
  Img.Free;
  SectorsImg.Free;
  inherited;
end;

procedure TSectorField.DrawSectors;
var
  X, Y: Integer;
  DX, DY: Integer;
  Rct: TRect;
begin
  if Assigned(FOnDrawSector)then
    for Y:=0 to HorSectorCount-1 do
      for X:=0 to VertSectorCount-1 do
      begin
        Rct:=GetSectorRect(Point(X, Y));
        DX:=0;
        DY:=0;
        if not AutoSize then
        begin
          DX:=Width div 2 - SectorsImg.Width div 2;
          DY:=Height div 2 - SectorsImg.Height div 2;
        end;
        Rct:=Rect(Rct.Left+DX, Rct.Top+DY, Rct.Right+DX, Rct.Bottom+DY);
        FOnDrawSector(Self, Img.Canvas, Point(X, Y), Rct);
      end;
end;

procedure TSectorField.DrawSectorsImg;
var
  X, Y: Integer;
  W, H: Integer;
begin
  W:=VertSectorCount*VertSectorSize;
  H:=HorSectorCount*HorSectorSize;
  SectorsImg.Width:=W+1;
  SectorsImg.Height:=H+1;
  with SectorsImg.Canvas do
  begin
    Brush.Color:=1;
    FillRect(ClipRect);
    Pen.Assign(FPen);
    for X:=0 to VertSectorCount do
    begin
      MoveTo(X*VertSectorSize, 0);
      LineTo(X*VertSectorSize, H);
    end;
    for Y:=0 to HorSectorCount do
    begin
      MoveTo(0, Y*HorSectorSize);
      LineTo(W, Y*HorSectorSize);
    end;
  end;
end;

function TSectorField.GetSectorRect(Sector: TPoint): TRect;
begin
  Result:=Rect(Sector.X*FVertSectorSize+1, Sector.Y*FHorSectorSize+1,
    (Sector.X+1)*FVertSectorSize, (Sector.Y+1)*FHorSectorSize);
end;

procedure TSectorField.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  Pt: TPoint;
begin
  inherited;
  if Assigned(FOnSectorDown)then
  begin
    Pt:=SectorAtPos(X, Y);
    if Pt.X<>-1 then
      FOnSectorDown(Self, Button, Shift, Pt.X, Pt.Y);
  end;
end;

procedure TSectorField.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Pt: TPoint;
begin
  inherited;
  if Assigned(FOnSectorMove)then
  begin
    Pt:=SectorAtPos(X, Y);
    if Pt.X<>-1 then
      FOnSectorMove(Self, Shift, Pt.X, Pt.Y);
  end;
end;

procedure TSectorField.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
var
  Pt: TPoint;
begin
  inherited;
  if Assigned(FOnSectorUp)then
  begin
    Pt:=SectorAtPos(X, Y);
    if Pt.X<>-1 then
      FOnSectorUp(Self, Button, Shift, Pt.X, Pt.Y);
  end;
end;

procedure TSectorField.Paint;
var
  X, Y: Integer;
begin
  inherited;
  Img.Width:=Width;
  Img.Height:=Height;
  with Img.Canvas do
  begin
    Brush.Color:=Color;
    FillRect(ClientRect);
    X:=0;
    Y:=0;
    if FDrawImageStyle=diCenter then
    begin
      X:=Width div 2 - FBackImage.Width div 2;
      Y:=Height div 2 - FBackImage.Height div 2;
    end;
    if FDrawImageStyle=diStretch then
      StretchDraw(ClientRect, FBackImage) else
        Draw(X, Y, FBackImage);
    if AutoSize then
      Draw(0, 0, SectorsImg)else
      begin
        X:=Width div 2 - SectorsImg.Width div 2;
        Y:=Height div 2 - SectorsImg.Height div 2;
        Draw(X, Y, SectorsImg);
      end;
  end;
  DrawSectors;
  Canvas.Draw(0, 0, Img);
end;

procedure TSectorField.PenChange(Sender: TObject);
begin
  DrawSectorsImg;
  Paint;
end;

procedure TSectorField.Redraw;
begin
  Paint;
end;

procedure TSectorField.Resize;
begin
  inherited;
  if AutoSize then Size;
end;

function TSectorField.SectorAtPos(X, Y: Integer): TPoint;
var
  DX, DY: Integer;
begin
  Result:=Point(-1, -1);
  if (VertSectorCount=0)or(HorSectorCount=0)then Exit;
  DX:=0;
  DY:=0;
  if not AutoSize then
  begin
    DX:=Width div 2 - SectorsImg.Width div 2;
    DY:=Height div 2 - SectorsImg.Height div 2;
  end;
  if (X-DX>=0)and(X-DX<VertSectorSize*VertSectorCount)and
  (Y-DY>=0)and(Y-DY<HorSectorSize*HorSectorCount)then
    Result:=Point((X-DX) div VertSectorSize, (Y-DY) div HorSectorSize);
end;

procedure TSectorField.SetAutoSize(Value: Boolean);
begin
  inherited;
  if AutoSize then Size;
end;

procedure TSectorField.SetBackImage(const Value: TBitmap);
begin
  FBackImage.Assign(Value);
end;

procedure TSectorField.SetDrawImageStyle(const Value: TDrawImgStyle);
begin
  FDrawImageStyle := Value;
  Paint;
end;

procedure TSectorField.SetHorSectorCount(const Value: Integer);
begin
  FHorSectorCount := Value;
  DrawSectorsImg;
  if AutoSize then Size else
    Paint;
end;

procedure TSectorField.SetHorSectorSize(const Value: Integer);
begin
  FHorSectorSize := Value;
  DrawSectorsImg;
  if AutoSize then Size else
    Paint;
end;

procedure TSectorField.SetPen(const Value: TPen);
begin
  FPen.Assign(Value);
  Paint;
end;

procedure TSectorField.SetVertSectorCount(const Value: Integer);
begin
  FVertSectorCount := Value;
  DrawSectorsImg;
  if AutoSize then Size else
    Paint;
end;

procedure TSectorField.SetVertSectorSize(const Value: Integer);
begin
  FVertSectorSize := Value;
  DrawSectorsImg;
  if AutoSize then Size else
    Paint;
end;

procedure TSectorField.Size;
begin
  Width:=VertSectorSize*VertSectorCount+1;
  Height:=HorSectorSize*HorSectorCount+1;
end;

{ TDisplayItems }

procedure TDisplayItems.Add(NewItem: TDisplayItem);
begin
  if NewItem<>nil then
  begin
    NewItem.OnChange:=FOnCallChange;
    NewItem.List:=List;
    List.Add(NewItem);
  end;
  DoChange;
end;

constructor TDisplayItems.Create;
begin
  List:=TList.Create;
  FOnCallChange:=nil;
end;

procedure TDisplayItems.Delete(Index: Integer);
var
  Obj: TDisplayItem;
begin
  if ItemExists(Index)then
  begin
    Obj:=List[Index];
    List.Delete(Index);
    if Obj<>nil then
    begin
      Obj.OnChange:=nil;
      Obj.List:=nil;
    end;
  end;
end;

destructor TDisplayItems.Destroy;
begin
  while List.Count<>0 do
    Delete(0);
  List.Free;
  inherited;
end;

procedure TDisplayItems.DoChange;
begin
  if Assigned(FOnCallChange)then
    FOnCallChange(Self);
end;

function TDisplayItems.GetCount: Integer;
begin
  Result:=List.Count;
end;

function TDisplayItems.GetItems(Index: Integer): TDisplayItem;
begin
  Result:=nil;
  if ItemExists(Index)then
    Result:=List[Index];
end;

function TDisplayItems.IndexOf(Item: TDisplayItem): Integer;
begin
  Result:=List.IndexOf(Item);
end;

function TDisplayItems.IndexOfName(ItemName: string): Integer;
var
  I: Integer;
  Item: TDisplayItem;
begin
  Result:=-1;
  for I:=0 to List.Count-1 do
  begin
    Item:=List[I];
    if Item<>nil then
      if Item.Name=ItemName then
      begin
        Result:=I;
        Exit;
      end;
  end;
end;

function TDisplayItems.ItemExists(Index: Integer): Boolean;
begin
  Result:=(Index>=0)and(Index<=List.Count-1);
end;

procedure TDisplayItems.Move(FirstIndex, SecondIndex: Integer);
begin
  if ItemExists(FirstIndex)and ItemExists(SecondIndex)then
    List.Move(FirstIndex, SecondIndex);
end;

procedure TDisplayItems.SetItems(Index: Integer;
  const Value: TDisplayItem);
var
  Item: TDisplayItem;
begin
  if ItemExists(Index)then
  begin
    Item:=List[Index];
    Item.Assign(Value);
  end;
end;

procedure TDisplayItems.SetOnCallChange(const Value: TNotifyEvent);
var
  I: Integer;
  Item: TDisplayItem;
begin
  FOnCallChange := Value;
  for I:=0 to List.Count-1 do
  begin
    Item:=List[I];
    Item.FOnChange:=FOnCallChange;
  end;
end;

procedure TDisplay.SetItems(const Value: TDisplayItems);
begin
  FItems.Assign(Value);
end;

end.
